// SPDX-License-Identifier: GPL-3.0
pragma solidity ^0.7.0;

//Declaración de contrato para realizar una compra y venta entre dos personas.
//Si cualesquiera de los dos aborta la compra, debe pagar una multa del 10%.
contract PurchaseandSell {

//Declaración de variables de estado globales para usar en el contrato.
    uint  inputvalue; //Valor del producto a comprar y vender.
    address payable  public seller; //Dirección de la cuenta del vendedor.
    address  payable public buyer; //Dirección de la cuenta del comprador.
    address  myAddress = address(this); //Dirección de la cuenta del contrato.
    uint public abort = 0; //Variable para indicar si se abortó la compra o la venta.
    uint   returnseller = 0; //Indica si el vendedor se retornó el dinero.
    uint   returnbuyer = 0;  //Indica si el comprador se retornó el dinero.  
    
//Declaración de un elemento enum para controlar el flujo de la aplicación.
    enum State { Created, Locked, Release, Inactive }

//Crea un avariable publica del tipo State, la variable se llama state con la
//primera letra minúscula para diferenciarla del nombre del enum. 
    State public state;

//Modificador de función para verificar condiciones en el flujo del programa.
    modifier condition(bool _condition) {
        require(_condition, "You are not fulfilling the conditions");
        _;
    }

//Modoficador para verificar que una función solo la ejecuta el comprador.
    modifier onlyBuyer() {
        require( msg.sender == buyer, "Only buyer can call this.");
        _;
    }

//Modificador para verificar que una función solo la ejecuta el vendedor.
    modifier onlySeller() {
        require(msg.sender == seller, "Only seller can call this.");
        _;
    }

//Modificador para verificar y controlar el flujo del código del programa.
    modifier inState(State _state) {
        require( state == _state,  "Invalid state.");
        _;
    }

//Modificador para verificar que el contrato tenga los fondos ingresados, tanto
//por el comprador como por el vendedor.
    modifier ContractFunds() {
        require( address(this).balance == 4*inputvalue,  "The contract has no funds");
        _;
    }

//Declaración de eventos para hacer seguimiento al flujo del código de programa.
    event Aborted(); //Abortado.
    event PurchaseConfirmed(); //Confirmación de la compra.
    event ItemReceived(); //Producto recepcionado.
    event SellerRefunded(); //Vendedor reembolsado.
  
//Función para habilitar la recepción de dinero por parte del contrato inteligente.
    receive() external payable { }

//Constructor para inicializar variables al compilar y ejecutar la aplicación
//en la interfaz de usuario de Remix.
    constructor(uint _valuepurchase)  {
        seller = msg.sender;
        inputvalue = _valuepurchase;        
        require((seller.balance) >= (4*inputvalue), "seller must have money.");        
        abort = 0;
    }
    
//Declaración de función para confirmar la compra, esta función la debe ejecutar
//el comprador, si la ejecuta el vendedor Solidity lanza error.
     function confirmPurchase() public inState(State.Created)  ContractFunds() {
        buyer = msg.sender;
        require(buyer != seller, "The buyer cannot be the seller");
        require(buyer.balance >= (2*inputvalue), "Buyer must have money.");
        emit PurchaseConfirmed();        
        state = State.Locked;
     }

//Función para abortar el proceso de compra y venta después de haber confirmado la
//la compra, sola la ejecuta el comprador, si la ejecuta debe pagar una multa
//al vendedor, el 10% del valor del producto, además debe aportar 2400 Gwei a la 
//cuenta del contrato (2400000000000 Wei).
    function AbortBuyer() public inState(State.Locked) condition((msg.sender == buyer)    
         && (msg.sender != seller)) payable  {
         require(msg.value==12400000000000,"Send the correct amount to the contract");
         state = State.Inactive;
         emit Aborted();
         seller.transfer((inputvalue/10));            
         abort = 1;
        }

//Función para que el comprador se reembolse el dinero que había aportado al contrato.
//Esta función la ejecuta solo el comprador después de haberse abortado el proceso
//de compra y venta por parte del comprador o del vendedor,
//también debe pagar una multa al contrato.
  function ReturntoBuyer() public inState(State.Inactive) condition((abort == 1 ) 
    && (msg.sender == buyer) && (msg.sender != seller) && returnbuyer == 0) payable{               
      if (returnseller == 0){
       require(msg.value == 2400000000000,"Send the correct amount to the contract");
             buyer.transfer((myAddress.balance)/2);
             returnbuyer = 1;
        } else if(returnseller == 1){
                 buyer.transfer((myAddress.balance)); 
                 returnbuyer = 1;
                 abort = 0;
               }                 
        }

//Función para abortar el proceso de compra y venta después de haber confirmado la
//la compra el comprador, sola la ejecuta el vendedor, si la ejecuta debe pagar una multa
//al comprador, el 10% del valor del producto, además debe aportar 2400 Gwei a la 
//cuenta del contrato (2400000000000 Wei).
  function AbortSeller() public inState(State.Locked) condition((msg.sender == seller)
    && (msg.sender != buyer) )  payable {
      require(msg.value == 12400000000000, "Send the correct amount to the contract");
         state = State.Inactive;
         emit Aborted();
         buyer.transfer(inputvalue/10 );            
         abort = 1;
        }

//Función para que el vendedor se reembolse el dinero que había aportado al contrato.
//Esta función la ejecuta solo el vendedor después de haberse abortado el proceso
//de compra y venta por parte del comprador o del vendedor, 
//también debe pagar una multa al contrato.
 function ReturntoSeller() public inState(State.Inactive) condition((abort == 1) 
   && (msg.sender == seller) && (msg.sender != buyer) && returnseller == 0) payable {               
     if (returnbuyer == 0){
       require(msg.value == 2400000000000, "Send the correct amount to the contract");
       seller.transfer((myAddress.balance)/2);
       returnseller = 1;
     } else if(returnbuyer == 1){
          seller.transfer((myAddress.balance));
          returnseller = 1;
          abort = 0;
       }                 
    }

//Función para confirmar la recepción del producto por parte del comprador.
//Solo la ejecuta el comprador, en su interior se cambia la variable state
//para evitar que la función vuelva a ser ejecutada por el
//comprador o el vendedor.
    function confirmReceived() public onlyBuyer inState(State.Locked)  {
        emit ItemReceived();        
        state = State.Release;
        buyer.transfer((myAddress.balance) - 3*inputvalue);
    }

//Función para pagar o reembolsar al vendedor, solamente la ejecuta el vendedor.
//en su interior se cambia la variable state para evitar que la función vuelva
//a ser ejecutada por el comprador o el vendedor.
    function refundSeller() public onlySeller inState(State.Release) {
        emit SellerRefunded();        
        state = State.Inactive;
        seller.transfer(myAddress.balance);        
    }

//Funciones para observar el balance de las cuentas del vendedor, el comprador
//y el balance de la cuenta del contrato. Esta funciones no son estrictamente
//necesarias en el desarrollo del código del programa, se instalan para 
//observar el desarrollo de las transacciones y como elemento didácto
//para verificar las explicaciones.
    function SeeBalanceseller()  view public returns (uint) {             
        return seller.balance;       
     }           
    function SeeBalancebuyer()  view public returns (uint) {            
         return buyer.balance;       
     }
    function SeeBalancecontract()  view public returns (uint) {            
         return address(this).balance;       
    }     

}