// SPDX-License-Identifier: GPL-3.0
pragma solidity ^0.8.11;

//Declaración de contrato para realizar una compra y venta entre dos personas.
contract dappbuyandsell {

//Declaración de variables de estado globales para usar en el contrato.
    uint  inputvalue; //Valor del producto a comprar y vender.
    address   payable public seller; //Dirección de la cuenta del vendedor.
    address  payable public buyer; //Dirección de la cuenta del comprador.
    address  myAddress = address(this); //Dirección de la cuenta del contrato.      
    uint bglobalseller; //Variable para observar el balance del vendedor.
    uint bglobalbuyer; //Variable para observar el balance del comprador.

    
    
//Declaración de un elemento enum para controlar el flujo de la aplicación.
    enum State { Created, Locked, Release, Inactive }

//Crea un avariable publica del tipo State, la variable se llama state con la
//primera letra minúscula para diferenciarla del nombre del enum. 
    State public state;    


//Modoficador para verificar que una función solo la ejecuta el comprador.
    modifier onlyBuyer() {
        require( msg.sender != seller, "Only buyer can call this.");
        _;
    }

//Modificador para verificar que una función solo la ejecuta el vendedor.
    modifier onlySeller() {
        require(msg.sender == seller, "Only seller can call this.");
        _;
    }

//Modificador para verificar y controlar el flujo del código del programa.
    modifier inState(State _state) {
        require( state == _state,  "Invalid state.");
        _;
    }

//Modificador para verificar que el contrato tenga los fondos ingresados, tanto
//por el comprador como por el vendedor.
    modifier ContractFunds() {
        require( address(this).balance == 4*inputvalue,  "The contract has no funds");
        _;
    }

//Declaración de eventos para hacer seguimiento al flujo del código de programa.    
    event PurchaseConfirmed(); //Confirmación de la compra.
    event ItemReceived(); //Producto recepcionado.
    event SellerRefunded(); //Vendedor reembolsado.
  
//Función para habilitar la recepción de dinero por parte del contrato inteligente.
    receive() external payable { }

//Constructor para inicializar variables al compilar y ejecutar la aplicación
//en la interfaz de usuario de Remix y en la interfaz de usuario de la Dapp.
    constructor()  {
        seller = payable(msg.sender);        
        bglobalseller = seller.balance;           
    }

//Función para ingresar el valor del producto a vender, solo la puede ejecutar 
 //el vendedor
    function ProductValue(uint _valuepurchase) public inState(State.Created) onlySeller()   {
        inputvalue = _valuepurchase;    
        require(inputvalue > 0, "Perform negotiation.");   
        require((seller.balance) >= (4*inputvalue), "seller must have money.");       
        bglobalseller = seller.balance;
    }


//Declaración de función para confirmar la compra, esta función la debe ejecutar
//el comprador, si la ejecuta el vendedor Solidity lanza error.
     function confirmPurchase() public inState(State.Created)  ContractFunds()   {           
        buyer = payable(msg.sender);     
        require(buyer != seller, "The buyer cannot be the seller");
        require(buyer.balance >= (2*inputvalue), "Buyer must have money.");
        emit PurchaseConfirmed();        
        state = State.Locked;
        bglobalbuyer = buyer.balance; 
        bglobalseller = seller.balance;       
      }

//Función para confirmar la recepción del producto por parte del comprador.
//Solo la ejecuta el comprador, en su interior se cambia la variable state
//para evitar que la función vuelva a ser ejecutada por el
//comprador o el vendedor.
    function confirmReceived() public onlyBuyer inState(State.Locked)  {
            emit ItemReceived();        
            state = State.Release;
            buyer.transfer((myAddress.balance) - 3*inputvalue);
            bglobalbuyer = buyer.balance; 
        }

//Función para pagar o reembolsar al vendedor, solamente la ejecuta el vendedor.
//en su interior se cambia la variable state para evitar que la función vuelva
//a ser ejecutada por el comprador o el vendedor.
    function refundSeller() public onlySeller inState(State.Release) {
            emit SellerRefunded();        
            state = State.Inactive;
            seller.transfer(myAddress.balance);   
            bglobalseller = seller.balance;     
        }

//Funciones para observar el balance de las cuentas del vendedor, el comprador
//y el balance de la cuenta del contrato. Esta funciones no son estrictamente
//necesarias en el desarrollo del código del programa, se instalan para 
//observar el desarrollo de las transacciones y como elemento didáctico
//para verificar las explicaciones. La primera función es para observar el 
//flujo del código del programa.
    function getState() public view returns (State) {
            return state;
      }  
      function SeeBalancebuyer()  view public returns (uint) {            
         return bglobalbuyer;       
     }
     function SeeBalanceseller()  view public returns (uint) {             
        return bglobalseller;       
     }
     function SeeBalancecontract()  view public returns (uint) {            
         return address(this).balance;       
    }
}