// SPDX-License-Identifier: GPL-3.0
pragma solidity ^0.7.0;

// Votaciones para elegir candidatos con la opción de delegar el voto una sola vez.
contract VotingContract {
    
// La estructura siguiente declara un objeto que se usa para almacenar
// variables de cada votante.
// Se usa en conjunto con el mapping voters (tabla para almacenar información).
// Esta estructura corresponde a cada votante.
    struct Voter {
        uint weight; //Peso del voto, vale 1 cuando se le asigna el derecho a votar.
        bool voted; //Es verdadero (true) si la persona ya ha votado.
        address delegate; // Persona a la que se delega el voto.
        uint vote; // Indice de la propuesta votada.
    }

//A continuación se declara un mapa o tabla de datos para almacenar
// la información de cada votante.
//El nombre es voters y funciona en conjunto con la estructura voter.
    mapping(address => Voter) public voters;

// En la siguiente estructura se almacenarán cada una de las propuestas
// a elegir por los votantes.
//Cada propuesta es una estructura diferente.
    struct Proposal {
        bytes32 name; // Nombre corto de 32 bytes.
        uint voteCount; // Número de votos acumulados.
    }

//En la línea de programa a continuación se define un vector (array)
// de tipo de la estructura proposal, para almacenar las propuestas
// a elegir, elvector es de tamaño fijo 3.
    Proposal[3] public typeproposals;
    
//En la línea de programa a continuación se declara una estructura
//para manejar la información de cada votante delegado.
    struct delegatedefinition {       
        uint youdelegate;//Entero (1 o 0) para controlar si un votante fue delegado.                
    }

//En la siguiente línea de programa se define un mapa o tabla de 
//datos para almacenar la información de cada votante delegado.
    mapping(address => delegatedefinition ) delegateassing;
   
//En la línea de programa a continuación se declara una estructura
//para manejar la información de cada votante delegante.
    struct delegatingdefinition {       
        address youdelegating; //Dirección del votante delegante.
         uint youdelegatingct; //Control de delegante.             
    }

//En la siguiente línea de programa se define un mapa o tabla de 
// datos para almacenar la información de cada votante delegante.
     mapping(address => delegatingdefinition ) delegatingassing;

//Declaración de vectores de tamaño fijo (arrays) y variables de estado globales.
    address  chairperson; //Presidente de la mesa y dueño de la dirección del contrato.    
    address[5] countsvoter; //Vector de direcciones de las cuentas de los votantes.
    uint[5]  identifiersvoter; //Vector de las identificaciones de los votantes.
    uint public accountant; //Contador de control de la cantidad de votantes.
    uint accountanttwo = 0; //Contador de votantes designados para votar.
    uint winningVoteCount = 0; //Contador de votos ganadores.
    uint winningProposalp; //Índice (número) de la propuesta ganadora.
    uint winningProptwo; //Índice (número) de la propuesta ganadora actualizado.   
    uint position; //Número de la propuesta ganadora corregido.
    address[2] countsdelegate; //Vector de cuentas de votantes delegados.
    address[2] countsdelegating; //Vector de cuentas de votantes delegantes.
    uint acountantthree; //Contador de votantes delegados.  
    uint[2] delegateID; //Vector de identificaciones de votantes delegados. 
    uint[2] delegatingID; //Vector de identificaciones de votantes delegantes.
    
        
//Función de recepción de ether para el contrato (opcional).
    receive() external payable { }

//Modificador de función para verificar que solo ejecuta un votante o
//que se pasó el límite de votantes solicitando votar.
     modifier onlyVoters() {
        require( msg.sender != chairperson && accountant < 5,
        "Only voted can voter o you reached the voter limit, please wait.");
        _;
    }

//Modificador de función para verificar que solo ejecuta un votante.
    modifier onlyVoterstwo() {
        require( msg.sender != chairperson, "Only voted can voter.");
        _;
    }

//Modificador de función para verificar que solo ejecuta un votante 
//o que se pasó el límite de votantes solicitando ser delegantes
// de una votación.
    modifier onlyVotersthree() {
        require( msg.sender != chairperson && acountantthree < 2,
         "Only voted can voter.");
        _;
    }

//Modificador de función para verificar los fondos del votante en su cuenta.
    modifier HasFunds() {
        require( msg.sender.balance > 0 , "Do not has funs.");
        _; 
    }

//Modificador de función para verificar que solo el presidente de 
//la mesa ejecuta la función donde se llame al modificador.
    modifier onlyChairperson() {
        require( msg.sender == chairperson , "Only chairperson can see.");
        _;
    }

//Modificador de función para verificar el derecho a votar del votante.
    modifier righttoVote() {
        require( voters[msg.sender].weight != 0 , "Has no right to vote");
        _;
    }

//Modificador de función para verificar que el votante no haya votado.
    modifier checkVote() {
        require( !voters[msg.sender].voted , "Already voted.");
        _;
    }

//Modificador de función para verificar que un votante no sea
// delegante por segunda vez.
    modifier verifyDelegating() {
        require( delegatingassing[msg.sender].youdelegatingct == 0,
         "you were already a delegating.");
        _;
    } 
    
//Ingreso de las propuestas a ser votadas, el ingreso es a través 
// del vector proposalNames.
//Los datos que se ingresan en el vector son de tipo bytes32.
//Los nombres de las propuestas quedan guardados en el vector
//typeproposals.
//Cada nombre se ingresa al final del vector proposals.
//typeproposals acepta el objeto Proposal porque él es como un método
// o instancia de éste.
//Se ingresan 3 nombres de propuestas, pero puede ampliarse
//cambiando la longitud del vector.
    constructor (bytes32[3] memory  proposalNames)  {
        chairperson = msg.sender;
        voters[chairperson].weight = 1;
        accountant = 0;         
        for (uint i = 0; i < proposalNames.length; i++) {
            typeproposals[i] = (Proposal({
            name: proposalNames[i],
            voteCount: 0
            }));
        }
    } 
   
//Función para ingresar los datos del votante, se ingresa una identificación
//personal e internamente se captura la dirección de la cuenta del votante.
    function voterData(uint _cedula)   public  onlyVoters() HasFunds()  {                   
        identifiersvoter[accountant] = _cedula; 
        countsvoter[accountant] = msg.sender;
        accountant += 1;         
    }

//Función para visualizar las identificaciones y las cuentas de los votantes.
//Solo puede ejecutar el presidente de la mesa.
    function seeVoters() view public onlyChairperson() returns(uint[5] memory,
     address[5] memory)  {        
          return (identifiersvoter, countsvoter );                     
    }

//Función para delegar el voto a un votante, solo la ejecuta un votante.
//El presidente de la mesa de votación no puede ejecutar esta función.
    function Delegated(uint _delegate, uint _delegating, address _countsdelegate ) public
      onlyVotersthree() HasFunds() verifyDelegating(){ 
            require( _countsdelegate != msg.sender && !voters[msg.sender].voted &&
               delegateassing[_countsdelegate].youdelegate == 0,
                "You must meet the delegate conditions.");  
            
            delegateID[acountantthree] = _delegate; 
            delegatingID[acountantthree] = _delegating;
            countsdelegate[acountantthree] = _countsdelegate;
            countsdelegating[acountantthree] = msg.sender;
            delegateassing[_countsdelegate].youdelegate = 1;
            delegatingassing[msg.sender].youdelegating = msg.sender;
            acountantthree += 1;        
    }

//Función para observar los votantes delegados.
//Solo es ejecutada por el presidente de la mesa de votación.
    function seeDelegated() view public onlyChairperson() returns(uint[2] memory,
      uint[2] memory, address[2] memory) {
         return (delegateID, delegatingID, countsdelegate); 
    }


//Función para asignar los derechos de votación.
//Sólo puede ser ejecutada por el presidente de la mesa (Chairperson).
//Se ingresan las cuentas de los votantes habilitados para votar.
//Las cuentas las ingresa el presidente de la mesa de votación.
//Las cuentas deben ser ingresadas a través del vector, voter.
    function giveRightToVote(address[5] memory adresvoter) public onlyChairperson() {
        
//En este for se deshabilita el delegante para que no pueda votar.
//También se asigna el registro de asignación como delegante.
        for (uint i = 0; i < countsdelegating.length; i++) {
            if(delegatingassing[countsdelegating[i]].youdelegating == countsdelegating[i]
             && countsdelegating[i] != address(0)){
                voters[countsdelegating[i]].voted = true;
                voters[countsdelegating[i]].weight = 1;
                delegatingassing[countsdelegating[i]].youdelegatingct = 1;
            }
            
        }

//En este for se asignan los derechos de votación de los votantes verificados
//e ingresados por el presidente de la mesa.
        for (uint i = 0; i < adresvoter.length; i++) {            
            if (delegateassing[adresvoter[i]].youdelegate == 1 &&
                 adresvoter[i] != address(0)) {
                   if(voters[adresvoter[i]].voted == true){
                       voters[adresvoter[i]].voted = false;
                       voters[adresvoter[i]].weight = 1;
                       accountanttwo += 1;
                   }else{
                       voters[adresvoter[i]].weight = 1;
                       accountanttwo += 1;
                   }
            } else if (voters[adresvoter[i]].voted == true && 
                voters[adresvoter[i]].weight != 0 && adresvoter[i] != address(0) ){
                    voters[adresvoter[i]].voted = true;
                    accountanttwo += 1;
            }else{
                    voters[adresvoter[i]].weight = 1; 
                     accountanttwo += 1;
            }

            if (accountanttwo == adresvoter.length){
                accountant = 0;
                accountanttwo = 0;
            }
        }

//En este for se analiza si los votantes delegados fueron habilitados para votar
//en caso contrario se actualizan las condiciones del delegante.
//También se actualizan los datos de ingreso del delegado y el delegante.
         for (uint i = 0; i < countsdelegate.length; i++) {
           if(voters[countsdelegate[i]].weight == 1 && countsdelegate[i] != address(0)&&
                voters[countsdelegate[i]].voted == false ){
                 voters[countsdelegate[i]].weight = 1;
             }else if(voters[countsdelegate[i]].weight == 0){
                 voters[countsdelegating[i]].voted = false;
                 voters[countsdelegating[i]].weight = 0;
                 delegatingassing[countsdelegating[i]].youdelegatingct = 0;
                 delegateassing[countsdelegate[i]].youdelegate = 0;
             }
                countsdelegating[i] = address(0);
                delegateID[i] = 0; 
                delegatingID[i] = 0;
                countsdelegate[i] = address(0);
                acountantthree = 0;
         }

//En estos ciclos for se actualizan los datos de ingreso de los votantes,
//tanto sus identificaciones como las direcciones de sus cuentas.
      for (uint i = 0; i < adresvoter.length; i++) {
            adresvoter[i] = 0x0000000000000000000000000000000000000000;//address(0)
       }        
      for (uint i = 0; i <  identifiersvoter.length; i++) {
            identifiersvoter[i] = 0;            
      }
      for (uint i = 0; i <  countsvoter.length; i++) {
       countsvoter[i] = 0x0000000000000000000000000000000000000000;//address(0)               
      }
    }

//Función para votar por alguna de las propuestas, inclusive los votantes
//que tienen un voto delegado.
//Si son tres propuestas se debe ingresar un número entre 1 y 3, al ingresar
//un número fuera de este rango Solidity arrojaría un error.
    function vote(uint proposal) public onlyVoterstwo() righttoVote() checkVote(){   
             
        voters[msg.sender].voted = true;
        voters[msg.sender].vote = proposal;
        position = proposal - 1;        
        typeproposals[position].voteCount += voters[msg.sender].weight;
        position = 0;
    }

//Función para Calcular la propuesta ganadora de acuerdo
//a todos los votos ingresados por los votantes.
//Esencialmente esto se calcula a través de una comparación matemática.
    function winningProposal() public onlyChairperson(){        
        for (uint p = 0; p < typeproposals.length; p++) {
            if (typeproposals[p].voteCount > winningVoteCount) {
                winningVoteCount = typeproposals[p].voteCount;
                winningProposalp = p;
                winningProptwo = p + 1;                
            }
        }        
    }

//Función para mostrar el número (índice) de la propuesta ganadora (primer dato) 
//y la cantidad de votos con los que se ganaría la votación (segundo dato).
//No se introducen caracteres string para evitar el uso de memoria
//y por tanto un mayor uso de Gas al ejecutar esta función.
    function showwinningProposal() view public onlyChairperson() returns(uint, uint){
            return(winningProptwo, winningVoteCount );
    }

//Función para mostrar el nombre de la propuesta ganadora.
//Esta función lo que hace es convertir una variable tipo bytes32
//en una variable tipo string.
    function winnerName()   public view onlyChairperson() returns (string memory )  {            
             bytes32  z = typeproposals[winningProposalp].name;
             bytes memory bytesString = new bytes(32);
             uint charCount = 0;
             for (uint j = 0; j < 32; j++) {
                byte char = byte(bytes32(uint(z) * 2 ** (8 * j)));
                if (char != 0) {
                    bytesString[charCount] = char;
                    charCount++;
                }
             }
            bytes memory bytesStringTrimmed = new bytes(charCount);
            for (uint j = 0; j < charCount; j++) {
                 bytesStringTrimmed[j] = bytesString[j];
            }
            return string(bytesStringTrimmed);
    }
}

